/*
-------------------------------------------------------------------
SUDOKU SENSEI 1.03: a Sudoku Explainer Engine
Copyright (C) 2005  Martin Knoblauch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA  02110-1301, USA.

Contact the author: comocomocomo AT users.sourceforge.net
Latest versions: http://sourceforge.net/projects/sudoku-sensei
-------------------------------------------------------------------

    demo2-explainer.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../../../sudoku-sensei/sudoku-pub.h"

#ifdef WINDOWS
#define SYSTEM_PAUSE system ("pause")
#else
#define SYSTEM_PAUSE
#endif

/*////////////////////////////////////////////////////////////// */

#define MAX_GRIDS 100
#define MAX_DIST 1000

typedef struct s_params
{
  FILE * in, * out;
  int blockWidth, blockHeight;
  int nGrids, verbose;
  int * offsetsX, * offsetsY;
  char samurai, guess, diagonals, wait, compact;
}
t_params;

int ParseParams (t_params * p, int argc, char *argv[]);
void FreeParams (t_params * p);

void PrintGroupInfo (t_board_p b, FILE * pf, int nGr);
void PrintCells (t_board_p b, FILE * pf, int nGr, int mCl);
void PrintSymbols (t_board_p b, FILE * pf, int mSy);
void PrintSymbolsCompact (t_board_p b, FILE * pf, int mSy);

const char *GroupTypes[] = { "bloque", "fila",
                             "columna", "diagonal" };

/*////////////////////////////////////////////////////////////// */

int main (int argc, char *argv[])
{
  t_params par;
  t_board_p b;
  e_state s;
  e_rule r;
  int c, N, nGr, nIGr, nCl, nSy, mCl, mSy;

  PrintGPL (stdout);  /* Show License */

  if (ParseParams (&par, argc, argv) != 0)
  {
    SYSTEM_PAUSE;
    return -1;
  }

  printf ("Ejecuta \"demo2-explainer -h\" para obtener ayuda.\n");
  printf ("Nivel de detalle: %d\n", par.verbose);

  if (par.samurai)
    b = ConstructSamuraiBoard (par.blockWidth,
                               par.blockHeight,
                               par.diagonals);
  else
    b = ConstructCustomBoard (par.blockWidth,
                              par.blockHeight,
                              par.diagonals,
                              par.nGrids,
                              par.offsetsX,
                              par.offsetsY);

  if (!b)
    fprintf (stderr, "ERROR: No hay suficiente memoria libre\n");
  else
  {
    if (par.in==stdin)
      printf ("Teclea el problema en crudo (puntos para "
              "casillas vacias: .1...3..etc\n");

    ReadBoard (b, par.in, NULL, NULL);
    s = GetState (b);

    if (par.in==stdin)
      getchar ();

    fprintf (par.out,
             "\nEstado: %s\n",
             s==unsolved ? "no resuelto" :
             s==impossible ? "imposible" :
             s==solved ? "resuelto" :
             "desconocido");

    PrintBoard (b, par.out, NULL, NULL);

    nCl = GetNumCells (b);
    nSy = GetNumSymbols (b);

    if (s==impossible)
      fprintf (par.out, "Problema imposible\n");
    else if (s==solved)
      fprintf (par.out, "Problema ya resuelto\n");
    else
      do
      {
        Solve (b, par.verbose, par.guess);

        r = GetLastRuleType (b);

        N = GetLastRuleLevel (b);
        nGr = GetLastRuleGroup (b);
        nIGr = GetLastRuleIntersectedGroup (b);
        mCl = GetLastRuleCellsMask (b);
        mSy = GetLastRuleSymbolsMask (b);

        if (par.verbose && (N>=par.verbose ||
                            r==guess ||
                            r==norule        ))
        {
          switch (r)
          {
            case symbols:
            /* FIRST RULE (part A): */
            /* N symbols are possible in only N cells, so */
            /* no other symbol can be possible in these cells */

              fprintf (par.out, "\nRegla 1A (N%d): ", N);
              PrintGroupInfo (b, par.out, nGr);
              fprintf (par.out, "\n ");
              PrintSymbols (b, par.out, mSy);
              fprintf (par.out, "posible(s) unicamente en");
              PrintCells (b, par.out, nGr, mCl);
              fprintf (par.out, "\n  --> ningun otro simbolo "
                                "puede estar en esa(s) casilla(s)\n");
              break;

            case cells:
            /* FIRST RULE (part B): */
            /* N cells have only a set of N different possible */
            /* symbols, so these are impossible in other cells */

              fprintf (par.out, "\nRegla 1B (N%d): ", N);
              PrintGroupInfo (b, par.out, nGr);
              fprintf (par.out, "\n ");
              PrintCells (b, par.out, nGr, mCl);
              fprintf (par.out, "solo puede(n) contener");
              PrintSymbols (b, par.out, mSy);
              fprintf (par.out, "\n  --> ese/os simbolo(s) no puede(n) "
                                "estar en otras casillas de el/la %s\n",
                                GroupTypes[GetGroupType(b,nGr)]);
              break;

            case inters:
            /* SECOND RULE: */
            /* A symbol is only possible in N cells, and they */
            /* all belong to the intersection with another */
            /* group, so it will not appear anywhere else in */
            /* that other group */

              fprintf (par.out, "\nRegla 2 (N%d): ", N);
              PrintGroupInfo (b, par.out, nGr);
              fprintf (par.out, "\n ");
              PrintSymbols (b, par.out, mSy);
              fprintf (par.out, "posible unicamente en la "
                                "interseccion con");
              PrintGroupInfo (b, par.out, nIGr);
              fprintf (par.out, "\n  --> ese simbolo no puede "
                                "estar en otras casillas de "
                                "este/a ultimo/a %s\n",
                                GroupTypes[GetGroupType(b,nIGr)]);
              break;

            case guess:

              fprintf (par.out, "\nAdivinando casilla: (f%dc%d)\n",
                       GetCellY (b, GetNumGuessedCells(b)-1)+1,
                       GetCellX (b, GetNumGuessedCells(b)-1)+1);
              break;

            case norule:
              fprintf (par.out,
                       "\nNo se puede hacer nada mas.\n");
          }

          fprintf (par.out, "\nCasilla(s) modificada(s):\n");

          for (c=0; c<nCl; c++)
            if (CellWasChangedByLastRule(b,c))
            {
              fprintf (par.out, "  (f%dc%d)-={",
                       GetCellY(b,c)+1, GetCellX(b,c)+1);

              PrintSymbolsCompact (b, par.out,
                  GetLastRemovedPossValuesOfCell(b,c));

              fputc ('}', par.out);
            }

          fprintf (par.out, "\n");
        }

        s = GetState (b);
        fprintf (par.out,
                 "\nEstado: %s\n",
                 s==unsolved ? "no resuelto" :
                 s==impossible ? "imposible" :
                 s==solved ? "resuelto" :
                 "desconocido");

        if (par.compact)
          PrintBoard (b, par.out, NULL, NULL);
        else
          PrintBoardPV (b, par.out, NULL, NULL);

        if (par.verbose && par.wait && par.out==stdout)
        {
          printf ("\nPulsa ENTER...");
          getchar ();
        }
      }
      while (s==unsolved && r!=norule);

    DestroyBoard (b);
  }

  FreeParams (&par);

  SYSTEM_PAUSE;
  return 0;
}

/*////////////////////////////////////////////////////////////// */

int ParseParams (t_params * p, int argc, char *argv[])
{
  int i, j;
  char ok, wrong, incompatible, notopen, nomem;

  p->in = stdin;                     /* Default parameters */
  p->out = stdout;
  p->blockWidth = p->blockHeight = 3;
  p->nGrids = 0;
  p->offsetsX = p->offsetsY = NULL;
  p->samurai = p->guess = p->diagonals = p->compact = FALSE;
  p->wait = TRUE;
  p->verbose = 1;

  ok = TRUE;
  wrong = incompatible = notopen = nomem = FALSE;

  for (i=1; i<argc && ok; i++)  /* Parse parameters */

    if (!strcmp(argv[i],"-samurai"))  /* Samurai board */
    {
      if (p->nGrids)
      {
        ok = FALSE;
        incompatible = TRUE;
        break;
      }

      p->samurai = TRUE;
    }
    else if (!strcmp(argv[i],"-grids"))  /* Custom board */
    {
      if (p->samurai)
      {
        ok = FALSE;
        incompatible = TRUE;
        break;
      }

      i ++;

      if (i>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->nGrids = -1;
      p->nGrids = atoi (argv[i]);

      if (p->nGrids<2 || p->nGrids>MAX_GRIDS)
      {
        p->nGrids = 0;
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->offsetsX = (int*) malloc (p->nGrids * sizeof(int));
      p->offsetsY = (int*) malloc (p->nGrids * sizeof(int));

      if (!p->offsetsX || !p->offsetsY)
      {
        ok = FALSE;
        nomem = TRUE;
        break;
      }

      for (j=0; j<p->nGrids; j++, i+=2)
      {
        if (i+2>=argc)
        {
          ok = FALSE;
          wrong = TRUE;
          break;
        }

        p->offsetsX[j] = p->offsetsY[j] = MAX_DIST+1;

        p->offsetsX[j] = atoi (argv[i+1]);
        p->offsetsY[j] = atoi (argv[i+2]);

        if (p->offsetsX[j]<-MAX_DIST ||
            p->offsetsX[j]>MAX_DIST ||
            p->offsetsY[j]<-MAX_DIST ||
            p->offsetsY[j]>MAX_DIST)
        {
          ok = FALSE;
          wrong = TRUE;
          break;
        }
      }

      if (!ok)
        break;
    }
    else if (!strcmp(argv[i],"-blocksize"))  /* Block size */
    {
      i ++;

      if (i+1>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->blockWidth = p->blockHeight = -1;

      p->blockWidth = atoi (argv[i]);
      p->blockHeight = atoi (argv[i+1]);

      if (p->blockWidth < 2 ||
          p->blockHeight < 2 ||
          p->blockWidth*p->blockHeight > 30)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      i ++;
    }
    else if (!strcmp(argv[i],"-if"))  /* Input file */
    {
      i ++;

      if (i>=argc ||
          !(p->in=fopen(argv[i],"rt")))
      {
        ok = FALSE;
        notopen = TRUE;
        break;
      }
    }
    else if (!strcmp(argv[i],"-of"))  /* Output file */
    {
      i ++;

      if (i>=argc ||
          !(p->out=fopen(argv[i],"wt")))
      {
        ok = FALSE;
        notopen = TRUE;
        break;
      }
    }
    else if (!strcmp(argv[i],"-guess"))  /* Guess */
    {
      p->guess = TRUE;
    }
    else if (!strcmp(argv[i],"-v"))  /* Verbose level */
    {
      i ++;

      if (i>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->verbose = -1;
      p->verbose = atoi (argv[i]);

      if (p->verbose<0)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }
    }
    else if (!strcmp(argv[i],"-diagonals"))  /* Diagonals */
    {
      p->diagonals = TRUE;
    }
    else if (!strcmp(argv[i],"-dontwait"))  /* Don't wait */
    {
      p->wait = FALSE;
    }
    else if (!strcmp(argv[i],"-compact"))  /* Old format */
    {
      p->compact = TRUE;
    }
    else if (!strcmp(argv[i],"-h"))  /* Help */
    {
      printf ("\n\tUSAGE:\n\tdemo2-explainer "
              "[parametros]\n\n"
              "\tPARAMETROS POSIBLES:\n"
              "\t  -h\t(mostrar esta ayuda)\n"
              "\t  -samurai\n"
              "\t  -grids n x1 y1 x2 y2... xn yn\n"
              "\t  -blocksize ancho alto\n"
              "\t  -diagonals\n"
              "\t  -guess\n"
              "\t  -v niveldedetalle\n"
              "\t  -if archivodeentrada\n"
              "\t  -of archivodesalida\n"
              "\t  -compact\n"
              "\t  -dontwait\n"
              "\n");

      FreeParams (p);
      return -1;
    }
    else
    {
      ok = FALSE;
      wrong = TRUE;
    }

  if (ok)
    return 0;

  if (notopen)
    perror ("ERROR abriendo archivo: ");
  else if (wrong)
    fprintf (stderr, "ERROR: Parametros incorrectos\n");
  else if (incompatible)
    fprintf (stderr, "ERROR: Parametros incompatibles\n");
  else if (nomem)
    fprintf (stderr, "ERROR: No hay suficiente memoria libre\n");

  FreeParams (p);
  return -1;
}

/*////////////////////////////////////////////////////////////// */

void FreeParams (t_params * p)
{
  if (p->offsetsX)
  {
    free (p->offsetsX);
    p->offsetsX = NULL;
  }

  if (p->offsetsY)
  {
    free (p->offsetsY);
    p->offsetsY = NULL;
  }

  if (p->in!=stdin)
  {
    fclose (p->in);
    p->in = stdin;
  }

  if (p->out!=stdout)
  {
    fclose (p->out);
    p->out = stdout;
  }
}

/*////////////////////////////////////////////////////////////// */

void PrintGroupInfo (t_board_p b, FILE * pf, int nGr)
{
  int i, x, y, xmin, xmax, ymin, ymax;

  xmin = xmax = GetCellX (b, GetCellOfGroup (b,0,nGr));
  ymin = ymax = GetCellY (b, GetCellOfGroup (b,0,nGr));

  for (i=1; i<GetNumSymbols(b); i++)
  {
    x = GetCellX (b, GetCellOfGroup (b,i,nGr));
    y = GetCellY (b, GetCellOfGroup (b,i,nGr));

    if (x>xmax) xmax = x;
    else if (x<xmin) xmin = x;

    if (y>ymax) ymax = y;
    else if (y<ymin) ymin = y;
  }

  fprintf (pf, " %s (f%dc%d)-(f%dc%d) ",
               GroupTypes[GetGroupType(b,nGr)],
               ymin+1, xmin+1, ymax+1, xmax+1);
}

/*////////////////////////////////////////////////////////////// */

void PrintCells (t_board_p b, FILE * pf, int nGr, int mCl)
{
  int i=0, n;

  n = GetNumSymbols (b);

  fprintf (pf, " casillas ");

  for (i=0; i<n; i++)
    if (mCl & (1<<i))
      fprintf (pf, "(f%dc%d) ",
                   GetCellY (b, GetCellOfGroup (b,i,nGr)) + 1,
                   GetCellX (b, GetCellOfGroup (b,i,nGr)) + 1);
}

/*////////////////////////////////////////////////////////////// */

void PrintSymbols (t_board_p b, FILE * pf, int mSy)
{
  int i=0, n;
  const char * symbols;

  n = GetNumSymbols (b);

  symbols = n <= 9 ?
            "123456789" :
            "0123456789ABCDEFHJKLMNPRTUVWYZ";

  fprintf (pf, " simbolos ");

  for (i=0; i<n; i++)
    if (mSy & (1<<i))
      fprintf (pf, "%c ", symbols[i]);
}

/*////////////////////////////////////////////////////////////// */

void PrintSymbolsCompact (t_board_p b, FILE * pf, int mSy)
{
  int i=0, n;
  const char * symbols;

  n = GetNumSymbols (b);

  symbols = n <= 9 ?
            "123456789" :
            "0123456789ABCDEFHJKLMNPRTUVWYZ";

  for (i=0; i<n; i++)
    if (mSy & (1<<i))
      fputc (symbols[i], pf);
}

/*////////////////////////////////////////////////////////////// */

