/*+----------------------------------------------+
  | Cippotto: WinDS Chip8 Emulator               |
  | (c) Francesco Lombardi 2006                  |
  +----------------------------------------------+
    Created for WinDS 2006 Compo
    http://www.gbawins.com
    http://itaprogaming.free.fr
  +----------------------------------------------+*/

#include "chip8wds.h"
#include <nds.h>

#define timer_count IPC->heartbeat

const unsigned char * Chip8Card[3] = {pong, tetris, brix};  /* A card with games            */
key_map Chip8Keys[3] = {pongKM, tetrisKM, brixKM};          /* An array containing keymaps  */

static byte get_timer_count(void)
{
  return timer_count;
}

bool isOnTop(Control *_this)
{
  list<Control* >::iterator i;
  i = --gWinS.GetSubControls()->end();
  if ( (*i) == _this)
    return true;
  else
    return false;
}

int run(void* __this)
{
  if (isOnTop((Control*)__this)) {
    chip8_running = 1;
    chip8_execute();
  }
  return 0;
}


/*************************************/
/* Chip8Console_Class implementation */
/*************************************/

/* Constructor */
Chip8Console_Class::Chip8Console_Class(const Bbox &b, int xdim, int ydim):
CtlGfxSurface(b, xdim, ydim)
{
  myMsg = gWinS.CreateCustomMessage();
}

/* _DrawForeground */
void Chip8Console_Class::_DrawForeground()
{
  this->CopyTo(upScreen, 0, 146);
  SendMessage(myMsg, this, 0);
}

/* HandleMessages */
void Chip8Console_Class::HandleMessages()
{
  WS_MSG msg;
  FORALLMESSAGES(msg) {
    if (msg.id == myMsg)
    {
      UpdateForeground();
    }
  }
  CtlGfxSurface::HandleMessages();
}

/* DrawFrame */
void Chip8Console_Class::DrawFrame()
{
  this->Fill(RGB2INT(10, 10, 10));
  this->DrawRectangle(Bbox(2, 2, 252, 42), WS_COLOR_WHITE);
  this->Printf(WS_DEFAULTS_SMALLFONT, 6, 6, WS_COLOR_WHITE, "Cippotto - WinDS Chip8 Emulator Ver.0.0.1");
  this->DrawHLine(6, 165, 14, WS_COLOR_WHITE);
}
//---------------------------------------------------------------------------


/************************************/
/* Chip8Screen_Class implementation */
/************************************/

/* Constructor */
Chip8Screen_Class::Chip8Screen_Class(const Bbox &b, int xdim, int ydim):
CtlGfxSurface(b, xdim, ydim)
{
  myMsg = gWinS.CreateCustomMessage();
}

/* _DrawForeground */
void Chip8Screen_Class::_DrawForeground()
{
  chip8_running = 1;
  chip8_execute();
  this->CopyTo(upScreen, 0, 18);
  SendMessage(myMsg, this, 0);
}

/* HandleMessages */
void Chip8Screen_Class::HandleMessages()
{
  WS_MSG msg;
  FORALLMESSAGES(msg) {
    if (msg.id == myMsg)
    {
      UpdateForeground();
    }
  }
  CtlGfxSurface::HandleMessages();
}
//---------------------------------------------------------------------------


/*********************************/
/* chip8wds_Class implementation */
/*********************************/

/* Constructor */
chip8wds_Class::chip8wds_Class():CtlWindow(Bbox(56, 16, 152, 40), "Cippotto", 0)
{
  /* Window controls creation */
  dlRomList = new CtlDropList(Bbox(3, 1, 140, 18));
  dlRomList->AddItem("pong");
  dlRomList->AddItem("tetris");
  dlRomList->AddItem("brix");
  dlRomList->AutoResizeX(true);
  dlRomList->AutoResizeY(true);

  /* Creating and centering emulator surface */
  EmuPaintBox = new Chip8Screen_Class(Bbox(0,0,0,0), 256, 128);
  EmuPaintBox->Fill(WS_COLOR_BLACK);

  /* Creating console surface */
  EmuConsole = new Chip8Console_Class(Bbox(0,0,0,0), 256, 50);

  /* Adding all subcontrols to window */
  AddSubControl(dlRomList);
  AddSubControl(EmuPaintBox);
  AddSubControl(EmuConsole);

  /* Some initializations */
  resetAll();
}

/* HandleMessages */
void chip8wds_Class::HandleMessages()
{
  WS_MSG msg;
  FORALLMESSAGES(msg)
  {
    switch(msg.id)
    {
      case WS_MSG_DROPDOWNLISTCHOICE:
        if( msg.data > -1 ) // None of the tabs are selected
        {
          chip8_reset();
          load();
          chip8_reset();
        }
      break;
    }
  }
  CtlWindow::HandleMessages();
}

/* update_display */
void chip8wds_Class::update_display (void)
{
  GfxSurface32k temp(256, 128);
  temp.Fill(WS_COLOR_BLACK);
  for (int y = 0; y < 32; ++y)
  {
    for (int x = 0; x < 64; x++)
    { /* Drawing routine: loops through all chip8 pixels and, if it is  */
      /* on, then draw a white block, else a black one.                 */
      //temp.Fill(Bbox(x*4, y*4, 4, 4), ((chip8_display[y * 64 + x] == 0xff) ? WS_COLOR_WHITE : WS_COLOR_BLACK));
      if (chip8_display[y * 64 + x] == 0xff)
        temp.Fill(Bbox(x*4, y*4, 4, 4), WS_COLOR_WHITE);
    }
  }
  temp.CopyTo((GfxSurface32k&)*EmuPaintBox);
}


/* interrupt                                                        */
/* Updates keyboard and display, sync emulation with hardware timer */
void chip8wds_Class::interrupt (void)
{
  check_keys();
  update_display();
}


/* check_keys */
void chip8wds_Class::check_keys(void)
{
  memset(chip8_keys, 0, sizeof(chip8_keys));
  c8ScanKeyPad();
  for (int i=0; i<16; i++) {
    if (c8KeysPressed() & (EmuKeys.KeyPad[i]))
      chip8_keys[i] = 1;
  }
}

/* resetAll */
void chip8wds_Class::resetAll(void)
{
  memset(chip8_mem, 0, sizeof(chip8_mem));  /* Zero-fills Chip8 memory */
  struct key_map EmuKeys =
    {
      "none",
      { KEY_UP,         KEY_LEFT,         KEY_RIGHT,          KEY_DOWN,
        KEY_A,          KEY_B,            KEY_X,              KEY_Y,
        KEY_L,          KEY_R,            KEY_SELECT,         KEY_START,
        KEY_R | KEY_UP, KEY_R | KEY_LEFT, KEY_R | KEY_RIGHT,  KEY_R | KEY_DOWN
      }
    };
  uperiod = 1;
  sync = 1;
  timer = IPC->heartbeat;
  chip8_iperiod=16;

  EmuConsole->DrawFrame();
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 6, 18, WS_COLOR_WHITE, "Made for WinDS 2006 Compo by Francesco Lombardi");
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 6, 26, WS_COLOR_WHITE, "Credits: MA, Francesco, LiraNuna, Bianca (<g>), NeoFlash Team,");
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 6, 34, WS_COLOR_WHITE, "        Marcel de Kogel (Vision8: CHIP8 emulation engine author)");
}

/* load */
int chip8wds_Class::load(void)
{
  memcpy(chip8_mem + 0x200, Chip8Card[dlRomList->GetSelectedItem()], sizeof(chip8_mem) - 0x200);
	EmuKeys = Chip8Keys[dlRomList->GetSelectedItem()];

  EmuConsole->DrawFrame();
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 6, 20, WS_COLOR_WHITE, "Game loaded: ");
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 60, 20, WS_COLOR_WHITE, Chip8Keys[dlRomList->GetSelectedItem()].RomName);
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 6, 30, WS_COLOR_WHITE, "Keys: ");

  char keystring[120] = "";
  for (int i = 0; i < 16; i++) {
    if (EmuKeys.KeyPad[i] > 0) {
      strcat(keystring, c8GetKeyName(EmuKeys.KeyPad[i]));
      strcat(keystring, "  ");
    }
  }
  EmuConsole->Printf(WS_DEFAULTS_SMALLFONT, 36, 30, WS_COLOR_WHITE, keystring);
  return 0;
}

/* Destructor */
chip8wds_Class::~chip8wds_Class()
{
  EmuPaintBox = NULL;
  delete EmuPaintBox;
  Chip8Emu = NULL;
}
